//
//crypt.c
//
//Interface for handling a crypto stream.
//
//
//-UserX 2001/11/16

#include <string.h>
#include "crypt.h"
#include "crypt-xt.h"
#include "crypt-dh.h"
#include "pipe.h"
#include "memiip.h"
#include "bignum.h"
#include "logger.h"
#include "striip.h"
#include "datablck.h"
#include "random.h"
#include "salt.h"

CryptStream blankcryptstream = BLANKCRYPTSTREAM;


CryptStream *csMake(void) {
	CryptStream *cs = memCopy(&blankcryptstream, sizeof(CryptStream), "CryptStream", NULL);//xiclone(&blankcryptstream, sizeof(CryptStream));
	LOGDEBUG(stringJoin("csMake:", ptrToString(cs)));
	cs->blowfishin = memAlloc(sizeof(BlowfishContext), "CryptStreamBlowfishContext", NULL);//ximalloc(sizeof(BlowfishContext));
	cs->blowfishout = memAlloc(sizeof(BlowfishContext), "CryptStreamBlowfishContext", NULL);//ximalloc(sizeof(BlowfishContext));
	return cs;
}

void csInit(CryptStream *cs, Pipe *thispipe) {//, Pipe *backpipe
	LOGDEBUG(stringJoin("csInit:", ptrToString(cs)));
	//cs->backPipe = backpipe;
	cs->thisPipe = thispipe;
}

void csFree(CryptStream *cs) {
	LOGDEBUG(stringJoin("csFree:", ptrToString(cs)));
	if(cs == NULL) {
		return;
	}
	bignumFree(cs->privkey);
	bignumFree(cs->localkey);
	bignumFree(cs->remotekey);
	bignumFree(cs->sharedkey);
	dblockFree(cs->insessionkey);
	dblockFree(cs->outsessionkey);
	bignumFree(cs->setremotekey);

	//ifree also does the clearing of memory. This is just being overly cautious.
	memset(cs->blowfishin, 0, sizeof(BlowfishContext));
	memset(cs->blowfishout, 0, sizeof(BlowfishContext));
	memFree(cs->blowfishin);//xifree(cs->blowfishin);
	memFree(cs->blowfishout);//xifree(cs->blowfishout);

	memcpy(cs, &blankcryptstream, sizeof(CryptStream));

	memFree(cs);//xifree(cs);
}

//xor rand into the next key position then rotates the session keys.
//will destroy the contents of rand
//individual key sizes are aasume to be the same size as rand
void csRotateSessionKey(DataBlock *key, DataBlock *rand) {
	int i;
	for(i = 0; i < rand->size; i++) {
		key->data[rand->size + i] ^= rand->data[i];
	}
	memmove(rand->data, key->data, rand->size);
	memmove(key->data, key->data + rand->size, key->size - rand->size);
	memmove(key->data + key->size - rand->size, rand->data, rand->size);

}

void csApplyCounter(uint8 *data, uint8 *counter) {
	int i;
	for(i = 0; i < BLOCK_LENGTH; i++) {
		data[i] ^= counter[i];
	}
	for(i = BLOCK_LENGTH1; i >= 0; i--) {
		if(++counter[i] != 0) {
			break;
		}
	}
}

void csDecode(CryptStream *cs, DataBlock *db) {
	int i;
	uint32 l, r;
	for(i = 0; i < db->size; i += BLOCK_LENGTH) {
		l = *((uint32 *)(&db->data[i + 0]));
		r = *((uint32 *)(&db->data[i + 4]));
#ifdef BIG__ENDIAN
		l = ENDIANSWAP32(l);
		r = ENDIANSWAP32(r);
#endif
		Blowfish_Decrypt(cs->blowfishin, (unsigned long *)&l, (unsigned long *)&r);
#ifdef BIG__ENDIAN
		l = ENDIANSWAP32(l);
		r = ENDIANSWAP32(r);
#endif
		*((uint32 *)(&db->data[i + 0])) = l;
		*((uint32 *)(&db->data[i + 4])) = r;
		csApplyCounter(db->data + i, cs->counterin);
	}
	//xteabDecode((uint32 *)db->data, db->size >> 2, (uint32 *)cs->insessionkey->data);
}

void csEncode(CryptStream *cs, DataBlock *db) {
	int i;
	uint32 l, r;
	for(i = 0; i < db->size; i += BLOCK_LENGTH) {
		csApplyCounter(db->data + i, cs->counterout);
		l = *((uint32 *)(&db->data[i + 0]));
		r = *((uint32 *)(&db->data[i + 4]));
#ifdef BIG__ENDIAN
		l = ENDIANSWAP32(l);
		r = ENDIANSWAP32(r);
#endif
		Blowfish_Encrypt(cs->blowfishout, (unsigned long *)&l, (unsigned long *)&r);
#ifdef BIG__ENDIAN
		l = ENDIANSWAP32(l);
		r = ENDIANSWAP32(r);
#endif
		*((uint32 *)(&db->data[i + 0])) = l;
		*((uint32 *)(&db->data[i + 4])) = r;
	}
	//xteabEncode((uint32 *)db->data, db->size >> XTEA_BYTES_PER_WORD_BITS, (uint32 *)cs->outsessionkey->data);
}

void csDecrypt(CryptStream *cs) {
	int insize;
	DataBlock *db;
	while(1) {
		if(cs->inNeeded == 0) {
			insize = BLOCK_LENGTH;
		} else if(cs->inNeeded == -1) {
			insize = KEY_LENGTH;
		} else {
			insize = (cs->inNeeded + BLOCK_LENGTH1) & (~BLOCK_LENGTH1);
		}
		if(cs->thisPipe->backPipe->inBuffer->size < insize) {
			return;
		}
		db = dblockCopyPart(cs->thisPipe->backPipe->inBuffer, 0, insize);
		saltAdd(db); //todo: this is probably not the best place to collect salt from.

		cs->thisPipe->backPipe->inBuffer = dblockDelete(cs->thisPipe->backPipe->inBuffer, 0, insize);
		csDecode(cs, db);
		if(cs->inNeeded == 0) {
			cs->inNeeded = db->data[0] + (db->data[1] << 8);
			if(cs->inNeeded == 0) {
				cs->inNeeded = -1;
			}
			LOGDEBUGTRAFFIC(stringJoin("csDecrypt: bytes waiting for ", intToHexString(cs->inNeeded)));
		} else if(cs->inNeeded == -1) {
			LOGDEBUG("csDecrypt: Got session key rotate:");
			LOGDEBUGKEY(stringJoin("key:", bufferToHexString(db->data, db->size)));
			csRotateSessionKey(cs->insessionkey, db);
			Blowfish_Init(cs->blowfishin, cs->insessionkey->data, KEY_LENGTH);
			cs->inNeeded = 0;
		} else {
			db->size = cs->inNeeded;
			cs->thisPipe->inBuffer = dblockAppendBlock(cs->thisPipe->inBuffer, db);
			LOGDEBUGTRAFFIC(stringJoin("csDecrypt: got data, length ", intToHexString(cs->inNeeded)));
			cs->inNeeded = 0;
		}
		dblockFree(db);
	}
	//int i;
	//for(i = 0; i < cs->thisPipe->backPipe->inBuffer->size; i++) {
	//	cs->thisPipe->backPipe->inBuffer->data[i] ^= cs->sessionkey->data[0];
	//}
	//cs->thisPipe->inBuffer = dblockAppendBlock(cs->thisPipe->inBuffer, cs->thisPipe->backPipe->inBuffer);
	//cs->thisPipe->backPipe->inBuffer->size = 0;
}

void csEncryptWrite(CryptStream *cs, DataBlock *db) {

	saltBuffer(db->data + db->size, (BLOCK_LENGTH - (db->size & BLOCK_LENGTH1)) & BLOCK_LENGTH1);

	db->size = (db->size + BLOCK_LENGTH1) & (~BLOCK_LENGTH1);

	csEncode(cs, db);
	cs->thisPipe->backPipe->outBuffer = dblockAppendBlock(cs->thisPipe->backPipe->outBuffer, db);

	saltAdd(db); //todo: this is probably not the best place to collect salt from.

	dblockFree(db);

}

void csEncrypt(CryptStream *cs) {
	DataBlock *db;
	DataBlock *dbr;
	int outsize;
	if(cs->thisPipe->outBuffer->size == 0) {
		return;
	}

	outsize = cs->thisPipe->outBuffer->size;

	if(outsize > 0xfff0) {
		outsize = 0xfff0;
	}
	LOGDEBUGTRAFFIC(stringJoin("csEncrypt: sending bytes ", intToHexString(outsize)));

	db = dblockMake(BLOCK_LENGTH, BLOCK_LENGTH, "csEncrypt");
	db->data[0] = outsize & 0xff;
	db->data[1] = (outsize >> 8) & 0xff;
	saltBuffer(db->data + 2, 6);

	csEncryptWrite(cs, db);

	db = dblockCopyPart(cs->thisPipe->outBuffer, 0, outsize);
	db = dblockResize(db, (outsize + BLOCK_LENGTH1) & (~BLOCK_LENGTH1));
	db->size = (outsize + BLOCK_LENGTH1) & (~BLOCK_LENGTH1);

	csEncryptWrite(cs, db);

	cs->thisPipe->outBuffer = dblockDelete(cs->thisPipe->outBuffer, 0, outsize);

	if(++cs->outCount >= WRITES_TILL_ROTATE) {
		cs->outCount = 0;

		db = dblockMake(BLOCK_LENGTH, BLOCK_LENGTH, "csEncryptRotate");
		db->data[0] = 0;
		db->data[1] = 0;

		csEncryptWrite(cs, db);

		dbr = randomDBlock(KEY_LENGTH);
		db = dblockCopy(dbr);

		csEncryptWrite(cs, db);

		LOGDEBUGTRAFFIC("csEncrypt: Sent session key rotate:");
		LOGDEBUGKEY(stringJoin("key:", bufferToHexString(dbr->data, dbr->size)));
		csRotateSessionKey(cs->outsessionkey, dbr);

		Blowfish_Init(cs->blowfishout, cs->outsessionkey->data, KEY_LENGTH);

		dblockFree(dbr);
	}



	//int i;
	//for(i = 0; i < cs->thisPipe->outBuffer->size; i++) {
	//	cs->thisPipe->outBuffer->data[i] ^= cs->sessionkey->data[0];
	//}
	//cs->thisPipe->backPipe->outBuffer = dblockAppendBlock(cs->thisPipe->backPipe->outBuffer, cs->thisPipe->outBuffer);
	//cs->thisPipe->outBuffer->size = 0;
}



void csInHandshake(CryptStream *cs) {
	int i;
	DataBlock *db;
	switch(cs->inShake) {
	case 1:
		if(cs->thisPipe->backPipe->inBuffer->size < DH_PUBLIC_KEY_LENGTH) {
			return;
		}
		db = dblockCopyPart(cs->thisPipe->backPipe->inBuffer, 0, DH_PUBLIC_KEY_LENGTH);
		cs->remotekey = bignumFromBuffer(db->data, DH_PUBLIC_KEY_LENGTH);
		LOGDEBUG(stringJoinMany("csInHandshake: ", ptrToString(cs), " got remote public key", NULL));
		LOGDEBUGKEY(stringJoin("remotekey:", bufferToHexString((uint8 *)cs->remotekey->data, DH_PUBLIC_KEY_LENGTH)));
		cs->thisPipe->backPipe->inBuffer = dblockDelete(cs->thisPipe->backPipe->inBuffer, 0, DH_PUBLIC_KEY_LENGTH);
		dblockFree(db);
		cs->inShake = 2;
		if(cs->setremotekey != NULL) {
			if(bignumCompare(cs->remotekey, cs->setremotekey) != 0) {
				LOGERROR("csInHandshake: remotekey is wrong!");
				LOGDEBUGKEY(stringJoin("remotekey:", bufferToHexString((uint8 *)cs->setremotekey->data, DH_PUBLIC_KEY_LENGTH)));
				cs->inShake = -1;
				cs->thisPipe->backPipe->status |= PSTATUS_CLOSED;
				cs->thisPipe->status |= PSTATUS_CLOSED;
				return;
			}
		}
	case 2:
		if(cs->localkey == NULL) {
			return;
		}
		cs->sharedkey = dhGenerateSharedKey(cs->privkey, cs->remotekey);
		randomAddEntropyClock(2);

		cs->insessionkey = dblockMake(DH_PUBLIC_KEY_LENGTH, DH_PUBLIC_KEY_LENGTH, "sessionkey");
		//cs->outsessionkey = dblockMake(DH_PUBLIC_KEY_LENGTH, DH_PUBLIC_KEY_LENGTH);
		LOGDEBUG(stringJoinMany("csInHandshake: ", ptrToString(cs), " made shared key", NULL));
		LOGDEBUGKEY(stringJoin("sharedkey:", bufferToHexString((uint8 *)cs->sharedkey->data, DH_PUBLIC_KEY_LENGTH)));

		cs->insessionkey->size = bignumToBuf(cs->insessionkey->data, cs->sharedkey, DH_PUBLIC_KEY_LENGTH);
		//bignumToBuf(cs->outsessionkey->data, cs->sharedkey, DH_PUBLIC_KEY_LENGTH);

		//todo: Consider using a hash to cut down keylength instead.
		for(i = 0; i < cs->insessionkey->size / 2; i++) {
			cs->insessionkey->data[i] ^= cs->insessionkey->data[i + cs->insessionkey->size / 2];
		}
		cs->insessionkey->size = cs->insessionkey->size / 2;
		for(i = 0; i < cs->insessionkey->size / 2; i++) {
			cs->insessionkey->data[i] ^= cs->insessionkey->data[i + cs->insessionkey->size / 2];
		}
		cs->insessionkey->size = cs->insessionkey->size / 2;

		cs->outsessionkey = dblockCopy(cs->insessionkey);

		Blowfish_Init(cs->blowfishin, cs->insessionkey->data, KEY_LENGTH);
		Blowfish_Init(cs->blowfishout, cs->outsessionkey->data, KEY_LENGTH);

		cs->inShake = 0;
		cs->outShake = 0;

		cs->inNeeded = 0;
	//default:
	}
}

void csOutHandshake(CryptStream *cs) {
	DataBlock *db;
	if(cs->outShake == 1) {
		if(cs->privkey == NULL) {
			cs->privkey = dhGeneratePrivKey();
		} else {
			LOGDEBUG("csOutHandshake: using premade private key");
		}
		cs->localkey = dhGeneratePubKey(cs->privkey);
		randomAddEntropyClock(2);

		LOGDEBUG(stringJoinMany("csOutHandshake: ", ptrToString(cs), " made private and local public key", NULL));
		LOGDEBUGKEY(stringJoin("privkey:", bufferToHexString((uint8 *)cs->privkey->data, DH_PRIVATE_KEY_LENGTH)));
		LOGDEBUGKEY(stringJoin("pubkey:", bufferToHexString((uint8 *)cs->localkey->data, DH_PUBLIC_KEY_LENGTH)));
		db = dblockMake(DH_PUBLIC_KEY_LENGTH, DH_PUBLIC_KEY_LENGTH, "pubkey");
		bignumToBuf(db->data, cs->localkey, DH_PUBLIC_KEY_LENGTH);
		cs->thisPipe->backPipe->outBuffer = dblockAppendBlock(cs->thisPipe->backPipe->outBuffer, db);
		dblockFree(db);
		cs->outShake = 2;
	}
	
}

void csRead(CryptStream *cs) {
	if(cs->inShake != 0) {
		csInHandshake(cs);
	} else {
		csDecrypt(cs);
	}
}

void csWrite(CryptStream *cs) {
	if(cs->outShake != 0) {
		csOutHandshake(cs);
	} else {
		csEncrypt(cs);
	}
}

