#!/usr/bin/perl
#
# iipconfgen.pl
# 
# A Configuration file generator for IIP
#   supports v1.1 only
# 
# Purpose:
#  To ease the setup of relay and server nodes.
#  It may also aid in the setup of user nodes.
# 
# Desired Capabilities:
#  For a user node:
#   get updated node.ref
#   create listen.ref 
#   create isproxy.ref
#   (possibly) show VFAQ
#  For a relay node:
#   generate (or load) relay keypair
#    get updated node.ref
#    create listen.ref
#    create isproxy.ini
#   export the relay's noderef
#  For a server node:
#   generate (or load) server keypair
#    create node.ref for server connections
#    create isproxy.ini for server connections
#
# Notes:
#  It should be possible to upgrade from user to relay easly
#  It should also be possible to change from relay to server
#   without affecting the external appearance of the relay
#  We should keep portablitily in mind in writeing this program.
#  Especialy to non-unix systems.
# 

my $version = '$Id: iipconfgen.pl,v 1.8 2002/04/11 19:44:37 chocolatezed Exp $';

use strict;
use Getopt::Std;

my %opts;
getopts('hdin:c:uU:rR:a:sS:gfk:x:GFK:X:b:B:l:y:z:', \%opts);

# CANCELLED make bestgess at hostname/ip
# TODO generate isproxy.ini for server 

if ($opts{h}) {
    print <<EOUSAGE ;
version: $version
usage: $0 [-h] [-din network.ref] [-urs]
    [-U[host][:port]] [-R[host][:port] [-a address]] [-S[host][:port]]
    [-gfk keyfile] [-x mynode.ref] [-GFK keyfile] [-X mynetwork.ref]
    [-b ./isproxy] [-B isproxy.ini] [-y node.ref] [-z listen.ref]

    -h		    show usage
    -d		    download new refs if available
    -i		    inform network of new public relay
    -n <filename>   use this network definition file
    -u		    configure as user node
    -r		    configure as relay node
    -s		    configure as server node
    -U [addr][:prt] set listen address and/or port for user
    -R [addr][:prt] set listen address and/or port for relay
    -a <address>    set externaly accessible address for relay
    -S [addr][:prt] set listen address and/or port for server
    -g		    generate new relay keypair
    -f		    force overwriteing of existing relay keyfile
    -k <filename>   import relay keys from this file, store keys in if -g
    -x <filename>   specify relay export file
    -G		    generate new server keypair
    -F		    force overwriteing of existing server keyfile
    -K <filename>   import server keys from this file, store keys in if -G
    -X <filename>   specify server export file
    -b <command>    isproxy to use for keypair generation
    -B <filename>   output to a different config file
    -l <filename>   extract keypairs from logfile, not stdout of isproxy
    -y <filename>   output to a different noderef file
    -z <filename>   output to a different listenref file

    To configure for a normal user node: $0 -ud
			     relay node: $0 -urdi
EOUSAGE
    exit;
}


#
# Sanitycheck arguments
#

$opts{U} and $opts{u} ||= 1;
$opts{R} and $opts{r} ||= 1;
$opts{S} and $opts{s} ||= 1;

my ($u_h, $u_p) = $opts{U} =~ /([\w.]+)?(?::(\d+))?/;
my ($r_h, $r_p) = $opts{R} =~ /([\w.]+)?(?::(\d+))?/;
my ($s_h, $s_p) = $opts{S} =~ /([\w.]+)?(?::(\d+))?/;

if (($opts{r}) and not $opts{a}) {
    die "You nead to specify address!\n";
}

unless ($opts{r} or $opts{s} or $opts{u}) {
    die "You nead to specify at least one of -u, -r or -s\n";
}

if ($opts{i}) {
    print "Automatic isproxy.ini configuration not implemented\n";
    print "You nead to configure your isproxy.ini to enable raw
connections.  (commenting out all the protocolgood and protocolbad fields
works)\n";
}

#
# Setup default configuration hashes
# 

my %files = (
    networkref	=> $opts{n} || "network.ref",
    noderef	=> $opts{y} || "node.ref",
    listenref	=> $opts{z} || "listen.ref",
    isproxyini	=> $opts{B} || "isproxy.ini",
    isproxybin	=> $opts{b} || "../isproxy",
    isproxylog  => $opts{l} || "iip.log",
);


my %network = (
    # entries used for 'meta' features
    userurl	=> "http://63.241.8.112/iip/crypto/node.ref",
    userfile	=> "usernode.ref",
    relayurl	=> "http://63.241.8.112/iip/node.ref",
    relayfile	=> "relaynode.ref",
    informurl	=> "http://63.241.8.112/iip/inform/inform.cgi",
    informfile	=> $opts{x}|| "mynode.ref",
    # entries used for generating ref files
    protocol	=> "spurt:backward:crypt:backward:raw",
    publickey	=> "<main pubkey as fallback>",
);

# Node.ref entries
my @node;

my %user = (
    name	=> "user",
    host	=> $u_h || "127.0.0.1",
    port	=> $u_p || "6667",
    protocol	=> $network{protocol},
);

my %relay = (
    name	=> "relay",
    address	=> $opts{a} || "somewhere.invisiblenet.net",
    host	=> $r_h || "0.0.0.0",
    port	=> $r_p || "1796",
    protocol	=> "crypt:raw",
    publickey	=> undef,
    privatekey	=> undef,
    exportfile	=> $opts{x} || "mynode.ref",
    keyfile	=> $opts{k} || undef, # "relay.key",
    keysource	=> $opts{g} ? "generate" : "import",
    keywrite	=> $opts{f} ? "force" : "keep",
);

my %server = (
    name	=> "server",
    host	=> $s_h || "127.0.0.1",
    port	=> $s_p || "6667",
    protocol	=> $network{protocol},
    publickey	=> undef,
    privatekey	=> undef,
    exportfile	=> $opts{X} || "mynetwork.ref",
    keyfile	=> $opts{K} || undef, # "server.key",
    keysource	=> $opts{G} ? "generate" : "import",
    keywrite	=> $opts{F} ? "force" : "keep",
);

#
# Fill in keypairs
# 

for ($opts{r} ? \%relay : () , $opts{s} ? \%server : ()) {
    print "- Processing $_->{name} keys -\n";
    my @rawdata; 
    if ($_->{keysource} eq "generate"
	    and $_->{keywrite} ne "force"
	    and -e $_->{keyfile}) {
	warn "Warning: keypair already generated for $_->{name}, useing that instead\n";
	$_->{keysource} = "import";
    }
    if ($_->{keysource} eq "generate") {
	print "Generating keys...\n";
	print "Saveing logfile...\n";
        rename $files{isproxylog}, $files{isproxylog} . "bak";
        unlink $files{isproxylog};
        if (not defined $opts{l}) {
            @rawdata = `$files{isproxybin} -K`;
            die "Error: bad isproxy specified\n" if $?;
        } else { # extract from logfile
            system("$files{isproxybin} -K");
            die "Error: bad isproxy specified\n" if $?;
            open my $logfile, "<", $files{isproxylog};
            @rawdata = <$logfile>;
            close $logfile;
        }
	print "Restoring logfile...\n";
        unlink $files{isproxylog};
        rename $files{isproxylog} . "bak", $files{isproxylog};
    } elsif ($_->{keysource} eq "import") {
	print "Importing keys...\n";	
	open my $keyfile, "<", $_->{keyfile};
	@rawdata = <$keyfile>;
	close $keyfile;
    } else {
	die "Error: invalid keysource '$_->{keysource}' for $_->{name} node config\n"
    }
    
    print "Extracting keys...\n";
    # Extract relavent lines from rawdata, and insert into hash
    # this has the nice feature that keyfiles generated by pipeing
    # 'isproxy -K' to a file will be properly read in.
    %$_ = (%$_, map {
	/^(?:\[.*?\]:)?(\w+)\s*key:(\w+)/; $1 ? ((lc $1 . 'key'), $2) : ()
    } @rawdata);
    
    unless ($_->{privatekey} and $_->{publickey}) {
        if ($_->{keysource} eq "import") {
            warn "Error importing $_->{name} keypair, retrying useing generation\n";
            $_->{keysource} = "generate";
            redo;
        }
        die "Error generating $_->{name} keypair\n";
    }
    
    # Save generated keypairs
    if ($_->{keysource} eq "generate" and defined $_->{keyfile}) {
	print "Saveing keys to $_->{keyfile}...\n";
	if ( $_->{keywrite} eq "force" or not -e $_->{keyfile}) {
	    open my $keyfile, ">", $_->{keyfile};
	    for my $key (qw(privatekey publickey)) {
		print $keyfile "${key}:$_->{$key}\n"
	    }
	} else {
	    warn "Warning: $_->{keyfile} already exists, cannot save keypair\n"
	}
    }
    print "Done processing $_->{name} keys\n";
}

#
# Read in configuration
# 


{
    my $optionpair = qr{^\s*(\w+)\s*=\s*(\S+)};
    my ($fh, $name);

    #
    # Read in network.ref
    #
    
    $name = $files{networkref};
    open $fh, "<", $name
	or die "Error: failed to open $name\n";

    while (<$fh>) {
	next if /^#/;
	next if /^$/;
	my ($key, $value) = /$optionpair/;
	next if !$key or !$value;
	$network{$key} = $value;
    }

    close ($fh);

    #
    # Read in node.ref
    #
    
    my $url = $network{ $opts{r} ? "relayurl" : "userurl" };
    $name = $network{ $opts{r} ? "relayfile" : "userfile" };
    
    if ($opts{d}) { # download latest node.ref
        use LWP::UserAgent;
        my $ua = LWP::UserAgent->new;
        my ($sver) = $version =~ /[^,]*v ([.\d]+)/;
        $ua->agent("IIPconfgen/$sver ");
        my $res = $ua->mirror($url, $name);
        print "download from $url => " . $res->status_line . "\n";
        if ($res->is_error()) {
            die "Download failed.  If you already have $name then you can try without the -d option";
        }
    }
    
    open $fh, "<", $name
	or die "Error: failed to open $name\n";

    my $node;
    while (<$fh>) {
	next if /^#/;
	next if /^$/;
	my ($key, $value) = /$optionpair/;
	next if !$key or !$value;
	
	$key eq "host" and do {
		if (ref $node) {
		   push @node, $node;
		   $node = {};
		}
	    };
	$node->{$key} = $value;
    }

    close ($fh);
}

#
# Generate configuration files
# 
# node.ref listen.ref isproxy.ini
#

if ($opts{u} or $opts{r} or $opts{s}) {
    my $fh;

    #
    # Generate Node.ref
    #

    open $fh, ">", $files{noderef}
	or die "Error: unable to open '$files{noderef}', $!";

    print "Writing '$files{noderef}'\n" if $opts{u} or $opts{r} or $opts{s};
    if ($opts{u} and !($opts{r} or $opts{s})) {
	print $fh "# user node.ref\n";
	print $fh "#\n";
	print $fh "networkid = $network{publickey}\n";
	print $fh "networkentry = $network{protocol}\n";
    }

    if (($opts{u} or $opts{r}) and !$opts{s}) {
	for my $node (@node) {
	    for (qw(host port protocol publickey)) {
		print $fh "$_ = $node->{$_}\n";
	    }
	}
    }

    if ($opts{s}) {
	print $fh "# server node.ref\n";
	print $fh "#\n";
	print $fh "networkid = $server{publickey}\n";
	print $fh "networksecret = $server{privatekey}\n";
	print $fh "networkentry = $server{protocol}\n";
	for (qw(host port)) {
	    print $fh "$_ = $server{$_}\n";
	}
	print $fh "special = exit\n";
    }

    close $fh;

    #
    # Generate listen.ref
    #


    if ($opts{u} or $opts{r}) {
        open $fh, ">", $files{listenref}
            or die "Error: unable to open '$files{listenref}', $!";

        print "Writing '$files{listenref}'\n" if $opts{u} or $opts{r};
        if ($opts{u}) {
            print $fh "# user configuration\n";
            for (qw(host port)) {
                print $fh "$_ = $user{$_}\n";
            }
            print $fh "special = entry\n";
            print $fh "# end user\n\n";
        }
        if ($opts{r}) {
            print $fh "# relay configuration\n";
            for (qw(host port protocol privatekey)) {
                print $fh "$_ = $relay{$_}\n";
            }
            print $fh "# end relay\n\n";
        }
    
        close $fh;
    }

    #
    # Nead to generate isproxy.ini for server
    #
}


#
# Export mynode.ref
#


if ($opts{r}) {
    open my $fh, ">", $relay{exportfile}
	or die "Error: unable to open '$relay{exportfile}', $!";
    
    print "Writing '$relay{exportfile}'\n";
    print $fh "# mynode.ref\n";
    print $fh "#\n";
    print $fh "host = $relay{address}\n";
    for (qw(port protocol publickey)) {
	print $fh "$_ = $relay{$_}\n";
    }
    
    close $fh;

    if ($opts{i}) { # inform?
        my $ua = LWP::UserAgent->new();
        my ($sver) = $version =~ /[^,]*v ([.\d]+)/;
        $ua->agent("IIPconfgen/$sver ");

        # adding POST stuff here
        my $res = $ua->post(
                $network{informurl},
                {
                    host => $relay{address},
                    port => $relay{port},
                    protocol => $relay{protocol},
                    publickey => $relay{publickey},
                },
            );
        print "upload to $network{informurl} => " . $res->status_line . "\n";
        if ($res->is_error()) {
            warn "Inform failed.\n";
        }
    }
}


#
# Export mynetwork.ref
#

if ($opts{s}) {
    open my $fh, ">", $server{exportfile}
	or die "Error: unable to open '$server{exportfile}', $!";

    print "Writing '$server{exportfile}'\n";
    print $fh <<EOSTUB ;
# mynetwork.ref
#
# These must be setup and filled in manualy
# prior to distributing the network.ref file
#
# userurl = http://0.0.0.0/iip/crypto/node.ref
# userfile = usernode.ref
# relayurl = http://0.0.0.0/iip/node.ref
# relayfile = relaynode.ref
# informurl = http://0.0.0.0/iip/inform/inform.cgi
# 
EOSTUB
    # entries used for generating ref files
    for (qw(protocol publickey)) {
	print $fh "$_ = $server{$_}\n";
    }
    
    close $fh;
}

__END__

=head1 NAME

iipconfgen - configuration generator for IIP

=head1 SYNOPSIS

See iipconfgen -h for complete option listing

Most options have sensible defaults.

IRC ports default to 6667
crypt:raw ports default to 1796
listenaddress for user nodes defaults to 127.0.0.1
listenaddress for public relays defaults to 0.0.0.0 (all addresses)

I<NOTE - isproxy is assumed to be in the parent dir>
You may have to tell it where to find isproxy

=head2 General Configuration

To configure as a user node:

    ./iipconfgen -du
    ./iipconfgen -du -U[listenaddress][:port]

To configure as a public relay node:

    ./iipconfgen -diru -a <public ip or hostname> -k <relay.key>
    ./iipconfgen -diru -a <public ip or hostname> -R[listenaddress][:port] -k <relay.key>

WARNING: Do NOT change the listen address of a public relay node unless
you know what you are doing.

=head2 Server Configuration

I<EXPERT USE ONLY - THIS WILL PROBABLY BREAK YOUR IIP INSTALL>

Server configuration is only neaded if you are going to run your own IIP
network.  Talk to the developers in #iip on the mail network or email
iip@hushmail.com if you are interested in setting up your own IIP network.

To configure as a server node add '-s' to your favorite public relay argument
list.  You will also nead to configure your isproxy.ini to enable raw
connections.  (commenting out all the protocolgood and protocolbad fields
works)

    ./iipconfgen -rs -a <public ip or hostname> -K server.key


=head1 DESCRIPTION

With the addition of relay and network keypairs to IIPv1, it was decided that
something was neaded to simplify configuration of public relay nodes.
This program attempts to deal with all resonable cases for configuring an
IIP node.

=head1 AUTHOR

Chocolate with a Zed

=head1 BUGS

Gessing the ip of the system has not been implimented yet.

=head1 SEE AlSO

The IIP homepage can be found at http://www.invisiblenet.net/iip/

=head1 Changelog

$Log: iipconfgen.pl,v $
Revision 1.8  2002/04/11 19:44:37  chocolatezed
Fixed some errors in the docs in the script
Fixed some server node only creation problems

Revision 1.7  2002/02/23 05:34:27  chocolatezed
Added inform capability to iipconfgen
Added perl inform cgi script

Revision 1.6  2002/02/17 07:36:19  chocolatezed
Implemented node.ref autoupdate feature

Revision 1.5  2002/02/11 09:06:29  chocolatezed
Added ability to read keypair from logfile
Fixed some of the arg checking logic

Revision 1.4  2002/02/10 06:49:50  chocolatezed
Added missing ; to line. Oops!

Revision 1.3  2002/02/10 06:28:41  chocolatezed
Fixed null keypair bug
Fixed server node.ref generation bug

Revision 1.2  2002/02/07 07:18:02  chocolatezed
Fixed typo in mynetwork.ref file generation
Added documentation about server node setup

Revision 1.1  2002/01/27 07:19:21  chocolatezed
Added iipconfgen


